/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   states_selftimed.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   April 5, 2006
 *
 *  Function        :   Output states visited in self-timed execution
 *
 *  History         :
 *      05-04-06    :   Initial version.
 *
 * $Id: states_selftimed.cc,v 1.1.1.1 2007/10/02 10:59:46 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include <math.h>  
#include <iostream>  
#include <assert.h>  
#include <time.h>  
#include <sys/resource.h>  

#include "../../analysis/statespace/statespace.h"
#include "../../base/algo/components.h"
#include "../../base/algo/repetition_vector.h"

#define TDTIME_MAX INT_MAX

#define max(a,b) ((a)>(b) ? (a) : (b))
#define min(a,b) ((a)<(b) ? (a) : (b))

typedef unsigned short TCnt;
typedef unsigned long TBufSize;
typedef double TDtime;

static TimedSDFgraph *g;
static SDFactor *outputActor;
static SDFtime SDF_MAX_TIME;
static TCnt SDF_OUT_ACTOR_REP_CNT;

static
SDFactor *findOutputActor(TimedSDFgraph *g, RepetitionVector repVec)
{
    int min = INT_MAX;
    SDFactor *a = NULL;
    
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        if (repVec[(*iter)->getId()] < min)
        {
            a = *iter;
            min = repVec[a->getId()];
        }
    }
    
    return a;
}

static
SDFtime getMaxExecTime(TimedSDFgraph *g)
{
    SDFtime max = 0;
    
    for (SDFactorsIter iter = g->actorsBegin(); iter != g->actorsEnd(); iter++)
    {
        TimedSDFactor *a = (TimedSDFactor*)(*iter);
        
        if (a->getExecutionTime() > max)
            max = a->getExecutionTime();
    }
    
    return max;
}

/******************************************************************************   * State  
 *****************************************************************************/
typedef struct _State  
{  
  TCnt **act_clk;
  TDtime glb_clk;  
  TBufSize *ch; 
} State;  

/**  
 * clearState ()  
 * The function sets the state to zero.  
 */  
static
void clearState(State &s)  
{
    for (uint i = 0; i < g->nrChannels(); i++)
        s.ch[i] = 0;
    
    s.glb_clk = 0;

    for (uint i = 0; i < g->nrActors(); i++)
        for (TCnt t = 0; t < SDF_MAX_TIME+1; t++)
            s.act_clk[i][t] = 0;
}  

/**  
 * equalStates ()  
 * The function compares to states and returns true if they are equal.  
 */  
static inline  
bool equalStates(const State &s1, const State &s2)  
{
    if (s1.glb_clk != s2.glb_clk)
        return false;
    
    for (uint i = 0; i < g->nrChannels(); i++)
        if (s1.ch[i] != s2.ch[i])
            return false;
    
    for (uint i = 0; i < g->nrActors(); i++)
        for (uint t = 0; t < SDF_MAX_TIME+1; t++)
            if (s1.act_clk[i][t] != s2.act_clk[i][t])
                return false;

    return true;
}  

/**  
 * copyState ()  
 * The function copies the state.  
 */  
static inline  
void copyState(State &to, const State &from)  
{  
    to.glb_clk = from.glb_clk;
    
    for (uint i = 0; i < g->nrChannels(); i++)
        to.ch[i] = from.ch[i];
    
    for (uint i = 0; i < g->nrActors(); i++)
        for (uint t = 0; t < SDF_MAX_TIME+1; t++)
            to.act_clk[i][t] = from.act_clk[i][t];
}

/**  
 * createState ()  
 * The function allocates memory for a new state.
 */  
static
void createState(State &s)  
{
    s.ch = new TBufSize [g->nrChannels()];
    
    s.act_clk = new TCnt* [g->nrActors()];
    for (uint i = 0; i < g->nrActors(); i++)
        s.act_clk[i] = new TCnt [SDF_MAX_TIME+1];
}  

/**  
 * destroyState ()  
 * The function deallocates all data structures of the state
 */  
static
void destroyState(State &s)  
{
    delete [] s.ch;
    
    for (uint i = 0; i < g->nrActors(); i++)
        delete [] s.act_clk[i];
    delete [] s.act_clk;
}  

/******************************************************************************
 * Stack
 *****************************************************************************/

/**  
 * StackPos  
 * Index into stack.  
 */  
typedef long long StackPos;  

/**  
 * stack  
 * The stack...  
 */  
static State *stack;  

/**  
 * stackPtr  
 * Index into stack of the first free space.  
 */  
static StackPos stackPtr;  
static StackPos maxStackPtr;  
static StackPos STACK_SIZE;  

/**
 * resizeStack ()
 * Enlarge the stack with a factor of 2.
 */
static
void resizeStack()
{
    StackPos newStackSize = 2 * STACK_SIZE;
    State *newStack;  

    // Allocate new stack
    newStack = new State [newStackSize];
    
    // Copy old stack to new stack
    for (StackPos p = 0; p < STACK_SIZE; p++)
        newStack[p] = stack[p];
    
    // Create state for new part of stack
    for (StackPos p = STACK_SIZE; p < newStackSize; p++)
        createState(newStack[p]);
    
    // Cleanup old stack
    delete [] stack;
    
    // Put new stack in place
    stack = newStack;
    STACK_SIZE = newStackSize;
}

/**  
 * createStack ()  
 * The function creates a stack.  
 */  
static
void createStack()  
{  
    stack = new State [STACK_SIZE];

    if (stack == NULL)
        throw CException("Failed creating stack.");  

    for (uint i = 0; i < STACK_SIZE; i++)
        createState(stack[i]);

    maxStackPtr = 0;  
}

/**  
 * destroyStack()  
 * The function destroys a stack.  
 */
static  
void destroyStack()  
{  
    for (uint i = 0; i < STACK_SIZE; i++)
        destroyState(stack[i]);

    delete [] stack;
}  

/**  
 * clearStack ()  
 * Reset the stack to an empty state.  
 */
static
void clearStack()  
{  
    stackPtr = 0;     
}  

/**  
 * popStack ()  
 * Remove last element from the stack.  
 */  
static inline  
void popStack()  
{  
    if (stackPtr > 0)  
	    stackPtr--;  
}  

/**  
 * pushStack ()  
 * Put an element on the stack.  
 */  
static inline  
void pushStack(const State &s)  
{  
    if (stackPtr >= STACK_SIZE)
    {
        cerr << "[INFO] buffy resizing stack." << endl;
        resizeStack();
    }
    
    copyState(stack[stackPtr], s);
    stackPtr++;

    if (stackPtr > maxStackPtr) maxStackPtr = stackPtr;
}  

/**  
 * topStack ()  
 * The function returns the last element of the stack.  
 */  
static inline  
State &topStack()  
{  
    return stack[stackPtr-1];  
}  
      /******************************************************************************
 * Hash
 *****************************************************************************/  
#define INVALID_HASH_KEY    NULL  
#define NO_SLOT_IN_HASH     -1  
	    
/**  
 * HashKey  
 *  
 */  
typedef unsigned long long HashKey;  

/**  
 * HashSlot  
 *  
 */  
typedef struct _HashSlot  
{  
	StackPos value;  
	struct _HashSlot *next;  
} HashSlot;  

/**  
 * hashTable  
 * The hash table...  
 */  
static HashSlot **hashTable;  
static HashKey HASH_TABLE_SIZE;

/**  
 * createHashTable ()  
 * The function constructs a hash table.  
 */  
static
void createHashTable()  
{  
	hashTable = new HashSlot* [HASH_TABLE_SIZE];

	if (hashTable == NULL)
        throw CException("Failed creating hash table.");  

	for (uint i = 0; i < HASH_TABLE_SIZE; i++)  
	    hashTable[i] = INVALID_HASH_KEY;  
}

/**  
 * destroyHashTable ()  
 * The function destoys a hash table.  
 */  
static
void destroyHashTable()  
{  
	HashSlot *s_cur, *s_next;  

	for (uint i = 0; i < HASH_TABLE_SIZE; i++)  
	{  
	    for (s_cur = hashTable[i]; s_cur != INVALID_HASH_KEY; s_cur = s_next)  
	    {  
	        s_next = s_cur->next;  
	        delete s_cur;  
	    }  
	}  

	delete [] hashTable;  
}  
	    
/**  
 * clearHashTable ()  
 * Resets the hash table to contain no keys.  
 */  
static
void clearHashTable()  
{  
	HashSlot *s_cur, *s_next;  

	for (uint i = 0; i < HASH_TABLE_SIZE; i++)  
	{  
	    for (s_cur = hashTable[i]; s_cur != INVALID_HASH_KEY; s_cur = s_next)  
	    {  
	        s_next = s_cur->next;
            delete s_cur;  
	    }  

	    hashTable[i] = INVALID_HASH_KEY;  
	}  
}  
	    
/**  
 * hash ()  
 * The hash function. It is a standard multiplication hash.  
 */  
static inline  
HashKey hash(const State &s)  
{  
	double key = 0;  
	long long key_part;  

    key_part = (long long)(s.glb_clk);
    key = key * 39164205.20662217 + double(key_part) * 0.6180339887;
    
    for (uint i = 0; i < g->nrActors(); i++)
    {
        for (TCnt t = 0; t < SDF_MAX_TIME+1; t++)
        {
            key_part = (long long)(s.act_clk[i][t]);
    	    key = key * 39164205.20662217 + double(key_part) * 0.6180339887;
        }
    }

    for (uint i = 0; i < g->nrChannels(); i++)
    {
        key_part = (long long)(s.ch[i]);
	    key = key * 39164205.20662217 + double(key_part) * 0.6180339887;
    }

	key = fmod(key, 1);  
	return (HashKey) floor(HASH_TABLE_SIZE * key);  
}  
	    
/**  
 * insertKeyHashTable ()  
 * The function inserts a (key, value) pair into the hash table.  
 */  
static inline  
void insertKeyHashTable(const HashKey key, const StackPos value)  
{  
	HashSlot *slot = new HashSlot;  
	slot->value = value;
	slot->next = hashTable[key];  

	// Insert the stack position in the hash table  
	hashTable[key] = slot;  
}  
	    
/**  
 * findVallueInHashTable ()  
 * The function returns the StackPos at which the given (key,state) pair is  
 * located or NO_SLOT_IN_HASH if unkown.  
 */  
static inline  
StackPos findValueInHashTable(const HashKey key, const State &s)  
{  
	HashSlot *slot;  

	for (slot = hashTable[key]; slot != INVALID_HASH_KEY; slot = slot->next)  
	{  
	    StackPos pos = slot->value;  

	    if (equalStates(stack[pos], s))  
	        return pos;  
	}  

	return NO_SLOT_IN_HASH;  
}  
      /******************************************************************************
 * SDF
 *****************************************************************************/      
/**  
 * outputIntervalCycle ()  
 * The function calculates the output interval of the states on the cycle. Its  
 * value is equal to the average time between two firings in the cycle.  
 */  
static inline  
TDtime outputIntervalCycle(const StackPos cyclePos)  
{  
	int nr_fire = 0;  
	TDtime time = 0;  

	// Check all state from stack till cycle complete  
	for (StackPos ptr = stackPtr -1; ptr >= cyclePos; ptr--)  
	{  
	    // Output actor fired?  
        if (stack[ptr].act_clk[outputActor->getId()][0] > 0)  
            nr_fire++;  

	    // Time between previous state  
	    time += stack[ptr].glb_clk;  
	}  

	return time/(TDtime)(nr_fire);  
}  
	    
/**  
 * storeState ()  
 * The function stores the given state at the stack and insert an entry  
 * into the hash table. Cycle detection is performed, as well as a  
 * check to verify wether a cycle is valid or not.  
 *  
 * Return values:  
 *      -1      - added state to stack  
 *      >= 0    - output interval of detected cycle  
 */  
static inline  
TDtime storeState(State &sdfState)  
{  
    // Calculate the key in the hash table  
    HashKey key = hash(sdfState);  

    // Check occurance in hash table  
    StackPos stackPos = findValueInHashTable(key, sdfState);  
    if (stackPos == NO_SLOT_IN_HASH)  
    {  
        // State not in hash table (i.e. state not seen before)  
        // Insert state in hash table  
        insertKeyHashTable(key, stackPtr);  

        // Insert state in stack  
        pushStack(sdfState);  
    }  
    else  
    {  
        // State has been visited before (i.e. found cycle)  
        return outputIntervalCycle(stackPos);  
    }  

    return -1;  
}  

#define GLB_CLK             sdfState.glb_clk  
#define ACT_CLK(a,t)        sdfState.act_clk[a][t]  
#define CH(c)               sdfState.ch[c]  

#define FIRE_ACT(a,t)       ACT_CLK(a,t)++;  
#define FIRE_ACT_END(a)     ACT_CLK(a,0)--;  
#define CH_TOKENS(c,n)      (CH(c) >= n)  
#define CONSUME(c,n)        CH(c) = CH(c) - n;
#define PRODUCE(c,n)        CH(c) = CH(c) + n;  
#define ACT_END_FIRE(a)     (ACT_CLK(a,0) != 0)  
#define ADVANCE_CLK         GLB_CLK = GLB_CLK + 1;  
#define NEXT_ITER           GLB_CLK = 0;  
#define LOWER_CLK(a)        for (TCnt t = 0; t < SDF_MAX_TIME; t++) \
                                ACT_CLK(a,t) = ACT_CLK(a,t+1); \
                            ACT_CLK(a,SDF_MAX_TIME) = 0;

static State sdfState;  
static State prevState;  

/**  
 * execSDFgraph()  
 * Execute the SDF graph till a deadlock is found or a recurrent state.  
 * The output interval (i.e. inverse of throughput) is returned.  
 */  
static
void execSDFgraph(ostream &out)  
{  
    TDtime output_interval;
    SDFtime nrClockSteps;
    bool actorFiringEnds;
    int repCnt = 0;  

    createState(sdfState);
    createState(prevState);  

    // Create initial state and add it to stack
    clearState(sdfState);  
    copyState(prevState, sdfState);  
    prevState.glb_clk = TDTIME_MAX;  

    // Initial tokens  
    for (SDFchannelsIter iter = g->channelsBegin(); 
            iter != g->channelsEnd(); iter++)
    {
        SDFchannel *ch = *iter;

        if (ch->getInitialTokens() != 0)
            PRODUCE(ch->getId(), ch->getInitialTokens());  
    }

    // Fire the actors  
    while (true)  
    {
        // Output current state
        out << "State: ((";
        for (uint i = 0; i < g->nrChannels(); i++)
        {
            out << sdfState.ch[i];
            
            if (i+1 < g->nrChannels())
                out << ",";
        }
        out << "),(";
        for (uint i = 0; i < g->nrActors(); i++)
        {
            out << "{";
            for (uint t = 0; t < SDF_MAX_TIME; t++)
            {
                if (sdfState.act_clk[i][t] != 0)
                {
                    for (uint j = 0; j < sdfState.act_clk[i][t]; j++)
                        out << t << ",";
                }
            }
            out << "}";
        }
        out << "))" << endl;
    
        // Finish actor firings
        out << "End:  ";
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            SDFactor *a = *iter;
    
            while (ACT_END_FIRE(a->getId()))  
            {
                if (outputActor->getId() == a->getId())
                {
                    repCnt++;
                    if (repCnt == SDF_OUT_ACTOR_REP_CNT)  
                    {  
                        // Add state to hash of visited states  
                        output_interval = storeState(sdfState);
                        if (output_interval != -1)
                        {
                            out << endl;
                            out << "Recurrent state found." << endl;
                            return;
                        }
	                    NEXT_ITER;  
                        repCnt = 0;  
                    }  

                }

                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::Out)
                        PRODUCE(ch->getId(), p->getRate());  
                }

                FIRE_ACT_END(a->getId());
                out << " " << a->getName();
            }  
        }
        out << endl;

        // Start actor firings  
        out << "Start:";
        for (SDFactorsIter iter = g->actorsBegin();
                iter != g->actorsEnd(); iter++)
        {
            TimedSDFactor *a = (TimedSDFactor*)*iter;

            while (true)
            {
                bool fire = true;
                
                // Check all input ports for tokens
                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::Out)
                        continue;

                    if (!CH_TOKENS(ch->getId(), p->getRate()))
                    {
                        fire = false;
                        break;
                    }
                }

                // Ready to fire?
                if (!fire)
                    break;

                // Consume tokens on each input  
                for (SDFportsIter pIter = a->portsBegin();
                        pIter != a->portsEnd(); pIter++)
                {
                    SDFport *p = *pIter;
                    SDFchannel *ch = p->getChannel();

                    if (p->getType() == SDFport::In)
                        CONSUME(ch->getId(), p->getRate());
                }

                // Fire the actor  
                FIRE_ACT(a->getId(), a->getExecutionTime());
                out << " " << a->getName();
            }
        }
        out << endl;
        
        // Progress maximal number of clock steps
        actorFiringEnds = false;
        nrClockSteps = 0;
        do
        {
            // Lower clocks  
            for (SDFactorsIter iter = g->actorsBegin();
                    iter != g->actorsEnd(); iter++)
            {
                SDFactor *a = *iter;
                LOWER_CLK(a->getId());
                
                if (sdfState.act_clk[a->getId()][0] != 0)
                    actorFiringEnds = true;
            }
            
            nrClockSteps++;
        } while (!actorFiringEnds);

	    // Check for progress (i.e. no deadlock)  
	    if (equalStates(prevState, sdfState))
        {
            out << "Deadlock" << endl;
 	        return;  
        }
        
        // Advance the global clock  
        GLB_CLK = GLB_CLK + nrClockSteps;
        out << "Clock: " << nrClockSteps << endl;

	    // Store state to check for progress  
	    copyState(prevState, sdfState);  
    }
    
    destroyState(sdfState);
    destroyState(prevState);
}  

/**
 * outputSelfTimedStateSpace ()
 * Output all states visited during the self-timed execution of the graph gr.
 */
extern
void outputSelfTimedStateSpace(TimedSDFgraph *gr, ostream &out,
        unsigned long long stackSz, unsigned long long hashSz)
{
    RepetitionVector repVec;

    // Copy arguments to globals
    g = gr;
	STACK_SIZE = stackSz;
    HASH_TABLE_SIZE = hashSz;

    // Compute repetition vector
    repVec = computeRepetitionVector(g);
    
    // Find the output actor and its repetition count
    outputActor = findOutputActor(g, repVec);
    SDF_OUT_ACTOR_REP_CNT = repVec[outputActor->getId()];

    // Find largest execution time of all actors
    SDF_MAX_TIME = getMaxExecTime(g);

    // Create hash and stack
    createStack(); clearStack();
	createHashTable(); clearHashTable();

    // Execute the graph
    execSDFgraph(out);

	// Cleanup
	destroyStack();
	destroyHashTable();
}
