/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   problem.h
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   October 4, 2006
 *
 *  Function        :   Schedule problem
 *
 *  History         :
 *      04-10-06    :   Initial version.
 *
 * $Id: problem.cc,v 1.2 2007/11/02 09:34:46 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include "problem.h"
#include "base/base.h"

/**
 * ScheduleProblem ()
 * Constructor.
 */
ScheduleProblem::ScheduleProblem(CString name, CNode *messagesNode,
        CNode *archGraphNode, CNode *systemUsageNode)
{
    // Name of schedule problem
    scheduleName = name;
    
    // Schedule function
    scheduler = NULL;
    
    // Schedule period
    schedulePeriod = CGetAttribute(messagesNode, "period");

    // Construct architecture graph
    archGraph = constructInterconnectGraph(archGraphNode, schedulePeriod);
    
    // Construct set of messages which must be scheduled
    constructMessages(messagesNode, archGraph, messages);
    
    // Set occupation of resource in the architecture
    setUsageInterconnectGraph(archGraph, systemUsageNode);
    
    // Problem not solved
    problemSolvedFlag = false;
}

/**
 * ~ScheduleProblem ()
 * Destructor.
 */
ScheduleProblem::~ScheduleProblem()
{
    // Cleanup architecture graph
    delete archGraph;
    
    // Cleanup messages
    for (MessagesIter iter = messages.begin(); iter != messages.end(); iter++)
        delete (*iter);
}

/**
 * solve ()
 * Solve a single scheduling problem.
 */
bool ScheduleProblem::solve()
{
    // A schedule function is set?
    if (scheduler == NULL)
        throw CException("[ERROR] No schedule function set.");
    
    // Find constraints from all related schedules which are already solved
    for (ScheduleSwitchesIter iter = scheduleSwitches.begin();
            iter != scheduleSwitches.end(); iter++)
    {
        ScheduleSwitch sw = *iter;
        
        // Is this schedule problem the source schedule in the switch?
        if (sw.from->getName() == getName())
            sw.to->constrainScheduleProblem(this);
        else
            sw.from->constrainScheduleProblem(this);
    }

    CTimer timer;
    startTimer(&timer);
    
    try
    {
        // Schedule this problem...
        problemSolvedFlag = scheduler->schedule();
    }
    catch (CException &e)
    {
        cerr << e;
        problemSolvedFlag = false;
    }    
    
    stopTimer(&timer);
    if (problemSolvedFlag)
    {
        logMsg("Solved '" + getName() + "'" 
                    + " (time: " + printTimer(&timer) + ")");
    }
    else
    {
        logMsg("Failed '" + getName() + "'" 
                    + " (time: " + printTimer(&timer) + ")");
    }

    
    return problemSolvedFlag;
}

/**
 * setScheduler ()
 * Set the given scheduler as the scheduler to be used in solving the problem.
 */
void ScheduleProblem::setScheduler(CostScheduler *s)
{
    scheduler = s;
    scheduler->setArchGraph(archGraph);
    scheduler->setMessages(messages);
    problemSolvedFlag = false;
}

/**
 * constrainScheduleProblem ()
 * The function marks slots on links of the architecture graph in the
 * schedule problem p which it uses in its own schedule.
 *
 * It is possible that the overlap between the schedules contains multiple
 * periods of the source and/or destination schedule.
 *
 */
void ScheduleProblem::constrainScheduleProblem(ScheduleProblem *p)
{
    bool isSrcSchedule = false, isDstSchedule = false;
    TTime timeOverlap, timeInSrc, timeInDst, periodSrc, periodDst, duration;    
    SlotTable slotsSrc = SlotTable(0), slotsDst = SlotTable(0);
    TTime overlap = 0;
    SlotReservations slots;
    Link *lSrc, *lDst;
    
    // Is this schedule problem not solved yet?
    if (!isProblemSolved())
        return;

    // Is this problem an empty problem (i.e. it takes no time)?
    if (schedulePeriod == 0)
        return;

    // Check wether this schedule is the source or destination in a switch with
    // the schedule problem p. Find also the maximal overlap which needs to be
    // considered between the two schedules.
    for (ScheduleSwitchesIter iter = scheduleSwitches.begin();
            iter != scheduleSwitches.end(); iter++)
    {
        ScheduleSwitch sw = *iter;
        
        // Is this schedule problem the source schedule in the switch?
        if (sw.to->getName() == p->getName() && sw.from->getName() == getName())
        {
            isSrcSchedule = true;
            overlap = overlap > sw.overlap ? overlap : sw.overlap;
        }
        
        // Is this schedule problem the destination schedule in the switch?
        if (sw.from->getName() == p->getName() && sw.to->getName() == getName())
        {
            isDstSchedule = true;
            overlap = overlap > sw.overlap ? overlap : sw.overlap;
        }
    }

    // Schedule period of src and dst schedule
    if (isSrcSchedule)
    {
        periodSrc = schedulePeriod;
        periodDst = p->getArchGraph()->getSlotTablePeriod();
    }
    else
    {
        periodSrc = p->getArchGraph()->getSlotTablePeriod();
        periodDst = schedulePeriod;
    }

    // Iterate over all links in the architecture
    for (LinksIter iter = archGraph->linksBegin();
            iter != archGraph->linksEnd(); iter++)
    {
        // Link l in src and dst scheduling problem
        if (isSrcSchedule)
        {
            lSrc = *iter;
            lDst = p->getArchGraph()->getLink(lSrc->getName());
        }
        else
        {
            lDst = *iter;
            lSrc = p->getArchGraph()->getLink(lDst->getName());
        }

        timeOverlap = 0;
        timeInSrc = periodSrc - (overlap % periodSrc);
        timeInDst = 0;
        while (timeOverlap < overlap)
        {
            // Find slot table in dst schedule which includes
            // timeInDst
            for (SlotTablesIter sIter = lDst->slotTableSeqBegin();
                sIter != lDst->slotTableSeqEnd(); sIter++)
            {
                slotsDst = *sIter;

                // timeInDst within this slot table?
                if (slotsDst.getStartTime() <= timeInDst 
                        && timeInDst <= slotsDst.getEndTime())
                {
                    break;
                }
            }

            // Find slot table in src schedule which includes
            // timeInSrc
            for (SlotTablesIter sIter = lSrc->slotTableSeqBegin();
                sIter != lSrc->slotTableSeqEnd(); sIter++)
            {
                slotsSrc = *sIter;

                // timeInSrc within this slot table?
                if (slotsSrc.getStartTime() <= timeInSrc 
                        && timeInSrc <= slotsSrc.getEndTime())
                {
                    break;
                }
            }

            // For how long is the table valid? Answer: Minimum of the
            // remaining length slot table of src and dst schedule
            if (slotsDst.getEndTime() - timeInDst 
                    < slotsSrc.getEndTime() - timeInSrc)
            {
                duration = slotsDst.getEndTime() - timeInDst;
            }
            else
            {
                duration = slotsSrc.getEndTime() - timeInSrc;
            }

            // Set slot occupation in src or dst
            if (isSrcSchedule)
            {
                slots = slotsSrc.getUsedSlotsOfSchedule();
                lDst->setUsedSlots(slots, timeInDst, timeInDst + duration);
            }
            else
            {
                slots = slotsDst.getUsedSlotsOfSchedule();
                lSrc->setUsedSlots(slots, timeInSrc, timeInSrc + duration);
            }

            // Next moment in time
            timeOverlap = timeOverlap + duration + 1;
            timeInSrc = (timeInSrc + duration + 1) % periodSrc;
            timeInDst = (timeInDst + duration + 1) % periodDst;
        }
    }
}

/**
 * setPreferedSlots ()
 * The function marks all slots in all links of the interconnect graph
 * as prefered slots (with the maximal preference level - UINT_MAX) when
 * this slot is used in the corresponding link of the interconnect graph g.
 */
void ScheduleProblem::setPreferedSlots(InterconnectGraph *g)
{
    Link *l;
    CSize slotTableSize;
    
    // NoC properties
    slotTableSize = archGraph->getSlotTableSize();
    
    // Iterate over all links in the interconnect graph
    for (LinksIter iter = g->linksBegin(); iter != g->linksEnd(); iter++)
    {
        Link *lg = *iter;
        
        // Find matching link in interconnect graph of this problem
        l = archGraph->getLink(lg->getId());
        
        // Mark any slot used in lg as a prefered slot in l
        for (uint i = 0; i < slotTableSize; i++)
        {
            if (lg->preferredSlots[i] > 0)
                l->preferredSlots[i] = UINT_MAX;
        }
    }
}

/**
 * setPreferedSlotsOnLinks ()
 * The function marks every slot in every link of the interconnect graph
 * in problem as prefered (with maximimal level - UINT_MAX) when this
 * slot is used in any of the scheduling problems solved so far. This scheduler
 * uses this information to minimize the number of newly allocated slots.
 * Leaving as many slots free as possible for other applications.
 */
void setPreferedSlotsOnLinks(ScheduleProblems &problems,
        ScheduleProblem *problem)
{
    // Iterate over the list of schedule problems and solve them one by one
    for (ScheduleProblemsIter iter = problems.begin();
            iter != problems.end(); iter++)
    {
        ScheduleProblem *p = *iter;
        
        // Problem p already solved?
        if (p->isProblemSolved())
        {
            problem->setPreferedSlots(p->getArchGraph());
        }
    }
}

/**
 * solveSetOfScheduleProblems ()
 * The function tries to find a valid schedule for each of the given
 * schedule problems. It considers relations between the problems. On success,
 * it return true. Else it returns false.
 */
bool solveSetOfScheduleProblems(ScheduleProblems &problems)
{
    bool success = true;

    // Iterate over the list of schedule problems and solve them one by one
    for (ScheduleProblemsIter iter = problems.begin();
            iter != problems.end(); iter++)
    {
        ScheduleProblem *p = *iter;
        
        // Set prefered slots in all links
        setPreferedSlotsOnLinks(problems, p);
        
        // Try to solve the problem
        if (!p->solve())
        {
            logError("Failed solving scheduling problem '"
                        + p->getName() + "'");
            success = false;
            
            // No chance of successfully solving the set of problems.
            break;
        }
    }
    
    return success;
}

/**
 * createNetworkMappingNode ()
 * Create an XML node which describes the mapping of messages send over
 * the interconnect to the resources in the interconnect.
 */
CNode *createNetworkMappingNode(ScheduleProblems &problems)
{
    CNode *networkNode, *messagesNode, *entityNode;
    ScheduleProblem *p;
    CostScheduler *s;
    
    // Network node
    networkNode = CNewNode("network");
    
    // Create node for each set of messages (i.e. each schedule problem)
    for (ScheduleProblemsIter iter = problems.begin();
            iter != problems.end(); iter++)
    {
        p = *iter;
        s = p->getScheduler();
        
        // Create messages node for a schedule
        messagesNode = CAddNode(networkNode, "messages");
        CAddAttribute(messagesNode, "name", p->getName());
        
        // Add schedule to messages node
        for (MessagesIter iterMsg = s->messagesBegin();
                iterMsg != s->messagesEnd(); iterMsg++)
        {
            Message *m = *iterMsg;
            SchedulingEntity *e = m->getSchedulingEntity();
            Route *r = e->getRoute();
            SlotReservations sr = e->getSlotReservations();
            CString route, slots;
            bool first;
            
            // Scheduling entity
            entityNode = CAddNode(messagesNode, "schedulingEntity");
            CAddAttribute(entityNode, "msg", m->getId());
            CAddAttribute(entityNode, "startTime", e->getStartTime());
            CAddAttribute(entityNode, "duration", e->getDuration());
            
            // Route of entity
            first = true;
            for (LinksIter iterL = r->linksBegin();
                    iterL != r->linksEnd(); iterL++)
            {
                Link *l = *iterL;
                
                if (!first)
                    route += ",";
                route += l->getName();
                first = false;
            }
            CAddAttribute(entityNode, "route", route);
            
            // Slot reservations of entity
            first = true;
            for (uint i = 0; i < sr.size(); i++)
            {
                if (sr[i])
                {
                    if (!first)
                        slots += ",";
                    slots += i;
                    first = false;
                }
            }
            CAddAttribute(entityNode, "slots", slots);
        }
    }
    
    return networkNode;
}

/**
 * outputNetworkBindingAsXML ()
 * Output the binding of messages to the interconnect as XML.
 */
void outputNetworkBindingAsXML(ScheduleProblems &problems, ostream &out)
{
    // SDF mapping node
    CNode *sdf3Node = CNewNode("sdf3");
    CAddAttribute(sdf3Node, "version", "1.0");
    CAddAttribute(sdf3Node, "type", "sdf");
    CAddAttribute(sdf3Node, "xmlns:xsi",
                            "http://www.w3.org/2001/XMLSchema-instance");
    CAddAttribute(sdf3Node, "xsi:noNamespaceSchemaLocation",
                            "http://www.es.ele.tue.nl/sdf3/xsd/sdf3-sdf.xsd");
    
    // Mapping node
    CNode *mappingNode = CAddNode(sdf3Node, "mapping");
    CAddNode(mappingNode, createNetworkMappingNode(problems));
    
    // Create document and save it
    CDoc *doc = CNewDoc(sdf3Node);
    CSaveFile(out, doc);
}

/**
 * createNetworkUsageNode ()
 * Create an XML node which describes the usage of the links in the
 * interconnect.
 */
CNode *createNetworkUsageNode(ScheduleProblems &problems)
{
    SlotReservations *slotReservations;
    CNode *networkNode, *linkNode;
    InterconnectGraph *archGraph;
    ScheduleProblem *p;
    uint slotTableSize;
    
    // Network node
    networkNode = CNewNode("network");

    // No problems given?
    if (problems.size() == 0)
        return networkNode;

    // Initially all slots in all links assumed free
    archGraph = problems.front()->getArchGraph();    
    slotTableSize = archGraph->getSlotTableSize();
    slotReservations = new SlotReservations [archGraph->nrLinks()];
    for (uint i = 0; i < archGraph->nrLinks(); i++)
    {
        slotReservations[i].resize(slotTableSize);
        for (uint j = 0; j < slotTableSize; j++)
            slotReservations[i][j] = false;
    }
    
    // Find for each link which slots are occupied at least one moment in time
    // in at least one architecture graph.
    for (ScheduleProblemsIter iter = problems.begin();
            iter != problems.end(); iter++)
    {
        p = *iter;
        archGraph = p->getArchGraph();
    
        for (LinksIter iterL = archGraph->linksBegin();
                iterL != archGraph->linksEnd(); iterL++)
        {
            Link *l = *iterL;
            
            for (SlotTablesIter iterS = l->slotTableSeqBegin();
                    iterS != l->slotTableSeqEnd(); iterS++)
            {
                SlotTable &s = *iterS;
                
                for (uint i = 0; i < slotTableSize; i++)
                {
                    if (!s.isSlotFree(i))
                        slotReservations[l->getId()][i] = true;
                }
            }
        }
    }    
    
    // Create a node for each link which indicates the occupied slots
    for (LinksIter iterL = archGraph->linksBegin();
            iterL != archGraph->linksEnd(); iterL++)
    {
        Link *l = *iterL;
        bool first = true;
        CString slots;

        // Slots used from the link
        for (uint i = 0; i < slotTableSize; i++)
        {
            if (slotReservations[l->getId()][i])
            {
                if (!first)
                    slots += ",";
                slots += i;
                first = false;
            }
        }

        // Using a slot from the link?
        if (!slots.empty())
        {
            // Add a link node to the network
            linkNode = CAddNode(networkNode, "link");
            CAddAttribute(linkNode, "name", l->getName());
            CAddAttribute(linkNode, "occupiedSlots", slots);
        }
    }
    
    return networkNode;
}

/**
 * outputNetworkUsageAsXML ()
 * Output the usage of the interconnect as XML.
 */
void outputNetworkUsageAsXML(ScheduleProblems &problems, ostream &out)
{
    // SDF mapping node
    CNode *sdf3Node = CNewNode("sdf3");
    CAddAttribute(sdf3Node, "version", "1.0");
    CAddAttribute(sdf3Node, "type", "sdf");
    CAddAttribute(sdf3Node, "xmlns:xsi",
                            "http://www.w3.org/2001/XMLSchema-instance");
    CAddAttribute(sdf3Node, "xsi:noNamespaceSchemaLocation",
                            "http://www.es.ele.tue.nl/sdf3/xsd/sdf3-sdf.xsd");
    
    // Mapping node
    CNode *usageNode = CAddNode(sdf3Node, "systemUsage");
    CAddNode(usageNode, createNetworkUsageNode(problems));
    
    // Create document and save it
    CDoc *doc = CNewDoc(sdf3Node);
    CSaveFile(out, doc);
}
