/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   list_scheduler.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   April 24, 2006
 *
 *  Function        :   Construct a list schedule for each processor in the
 *                      platform graph.
 *
 *  History         :
 *      24-04-06    :   Initial version.
 *
 * $Id: list_scheduler.cc,v 1.2 2007/11/02 09:34:46 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include "list_scheduler.h"
#include "../../base/algo/components.h"
#include "../../base/algo/repetition_vector.h"

/******************************************************************************   * Architecture platform
 *****************************************************************************/

/**
 * init ()
 * Initialize architecture platform.
 */
void SDFstateSpaceListScheduler::ArchPlatform::init(TimedSDFgraph *appGraph,
        PlatformGraph *archGraph)
{
    // Not application or architecture given?
    if (appGraph == NULL || archGraph == NULL)
    {
        nrTiles = 0;
        return;
    }
    
    // Number of tiles in architecture
    nrTiles = archGraph->nrTiles();

    // Resize vectors
    actorBinding.resize(appGraph->nrActors());
    schedule.resize(archGraph->nrTiles());
    tdmaSize.resize(archGraph->nrTiles());
    tdmaSlice.resize(archGraph->nrTiles());
    
    // Initialize actor bindings
    for (uint i = 0; i < appGraph->nrActors(); i++)
        actorBinding[i] = ACTOR_NOT_BOUND;

    // Convert architecture graph to architecture platform
    for (TilesIter iter = archGraph->tilesBegin();
            iter != archGraph->tilesEnd(); iter++)
    {
        Tile *t = *iter;
        Processor *p = t->getProcessor();

        // Tile contains processor?
        if (p != NULL)
        {
            // Set binding of actors to the tile containing this processor
            for (ComponentBindingsIter bIter = p->getActorBindings()->begin();
                    bIter != p->getActorBindings()->end(); bIter++)
            {
                CId actorId = (*bIter)->getComponent()->getId();
                actorBinding[actorId] = t->getId();
            }
            
            // Set the schedule for the actors in the tile
            schedule[t->getId()] = p->getSchedule();
            
            // TDMA wheel size
            tdmaSize[t->getId()] = p->getTimewheelSize();
            
            // TDMA time slice of application
            tdmaSlice[t->getId()] = p->getReservedTimeSlice();
        }
    }
}

/******************************************************************************
 * State
 *****************************************************************************/

/**
 * clear ()
 * The function sets the state to zero.
 */
void SDFstateSpaceListScheduler::TransitionSystem::State::clear()
{
    for (uint i = 0; i < actClk.size(); i++)
    {
        actClk[i].clear();
    }

    for (uint i = 0; i < ch.size(); i++)
    {
        ch[i] = 0;
    }
    
    glbClk = 0;
    
    for (uint i = 0; i < schedulePos.size(); i++)
    {
        schedulePos[i] = NULL;
        tdmaPos[i] = 0;
    }
}

/**
 * operator= ()
 * The function compares to states and returns true if they are equal.
 */
bool SDFstateSpaceListScheduler::TransitionSystem::State::operator==(
        const State &s)
{
    if (glbClk != s.glbClk)
        return false;
    
    for (uint i = 0; i < ch.size(); i++)
    {
        if (ch[i] != s.ch[i])
            return false;
    }
    
    for (uint i = 0; i < actClk.size(); i++)
    {
        if (actClk[i] != s.actClk[i])
            return false;
    }

    return true;
}

/**
 * print ()
 * Print the state to the supplied stream.
 */
void SDFstateSpaceListScheduler::TransitionSystem::State::print(ostream &out)
{
    out << "### State ###" << endl;

    for (uint i = 0; i < actClk.size(); i++)
    {
        out << "actClk[" << i << "] =";
        
        for (list<SDFtime>::const_iterator iter = actClk[i].begin();
                iter != actClk[i].end(); iter++)
        {
            out << " " << (*iter) << ", ";
        }
        
        out << endl;
    }

    for (uint i = 0; i < ch.size(); i++)
    {
        out << "ch[" << i << "] = " << ch[i] << endl;
    }

    out << "glbClk = " << glbClk << endl;
    
    for (uint i = 0; i < schedulePos.size(); i++)
    {
        out << "tile[" << i << "] = (";
        if (schedulePos[i] != NULL)
            out << schedulePos[i]->id;
        else
            out << "NULL";
        out << ", " << tdmaPos[i] << ")" << endl;
    }
}

/******************************************************************************
 * Transition system
 *****************************************************************************/

/**
 * initOutputActor ()
 * The function selects an actor to be used as output actor in the
 * state transition system.
 */
void SDFstateSpaceListScheduler::TransitionSystem::initOutputActor()
{
    RepetitionVector repVec;
    int min = INT_MAX;
    SDFactor *a = NULL;

    // Compute repetition vector
    repVec = computeRepetitionVector(appGraph);
    
    // Select actor with lowest entry in repetition vector as output actor
    for (SDFactorsIter iter = appGraph->actorsBegin();
            iter != appGraph->actorsEnd(); iter++)
    {
        if (repVec[(*iter)->getId()] < min)
        {
            a = *iter;
            min = repVec[a->getId()];
        }
    }
    
    // Set output actor and its repetition vector count
    outputActor = a;
    outputActorRepCnt = repVec[outputActor->getId()];
}

/**
 * initArchPlatform ()
 * The function initializes the architecture platform.
 */
void SDFstateSpaceListScheduler::TransitionSystem::initArchPlatform(
        TimedSDFgraph *g, PlatformGraph *ap)
{
    // Initialize the architecture platform
    archPlatform.init(g, ap);
    
    // Construct empty schedule for each processor
    for (uint i = 0; i < archPlatform.nrTiles; i++)
        archPlatform.schedule[i] = NULL;
}

/**
 * storeState ()
 * The function stores the state s on whenever s is not already in the
 * list of storedStates. When s is stored, the function returns true. When the
 * state s is already in the list, the state s is not stored. The function
 * returns false. The function always sets the pos variable to the position
 * where the state s is in the list.
 */
bool SDFstateSpaceListScheduler::TransitionSystem::storeState(State &s,
        StatesIter &pos)
{
    // Find state in the list of stored states
    for (StatesIter iter = storedStates.begin();
            iter != storedStates.end(); iter++)
    {
        State &x = *iter;
        
        // State s at position iter in the list?
        if (x == s)
        {
            pos = iter;
            return false;
        }
    }
    
    // State not found, store it at the end of the list
    storedStates.push_back(s);
    
    // Added state to the end of the list
    pos = storedStates.end();
    
    return true;
}

/**  
 * computeThroughput ()  
 * The function calculates the throughput of the states on the cycle. Its  
 * value is equal to the average number of firings of an actor per time unit.
 */  
TDtime SDFstateSpaceListScheduler::TransitionSystem::computeThroughput(
        const StatesIter cycleIter)  
{  
	int nr_fire = 0;
	TDtime time = 0;  

	// Check all state from stack till cycle complete  
	for (StatesIter iter = cycleIter; iter != storedStates.end(); iter++)
    {
        State &s = *iter;

        // Number of states in cycle is equal to number of iterations 
        // in the period
        nr_fire++;

	    // Time between previous state  
	    time += s.glbClk;
	}

	return (TDtime)(nr_fire)/(time);  
}

/******************************************************************************
 * SDF
 *****************************************************************************/

#define CH(c)               currentState.ch[c]
#define SCHEDULE_POS(p)     currentState.schedulePos[p]
#define TDMA_POS(p)         currentState.tdmaPos[p]

#define CH_TOKENS(c,n)      (CH(c) >= n)  
#define CONSUME(c,n)        CH(c) = CH(c) - n;
#define PRODUCE(c,n)        CH(c) = CH(c) + n;

#define CH_TOKENS_PREV(c,n) (previousState.ch[c] >= n)  

/**
 * actorReadyToFire ()
 * The function returns true when the actor is ready to fire in state
 * s. Else it returns false.
 */
bool SDFstateSpaceListScheduler::TransitionSystem::actorReadyToFire(SDFactor *a)
{
    // Actor bound to processor?
    if (archPlatform.actorBinding[a->getId()] != ACTOR_NOT_BOUND)
    {
        uint p = archPlatform.actorBinding[a->getId()];
        
        // Processor is not idle or actor is not first in readyList?
        if (!procIdle[p] || actorReadyList[p].empty() 
                || actorReadyList[p].front()->getId() != a->getId())
        {
            return false;
        }
    }
    
    // Check all input ports for tokens
    for (SDFportsIter iter = a->portsBegin(); iter != a->portsEnd(); iter++)
    {
        SDFport *p = *iter;
        SDFchannel *c = p->getChannel();
        
        // Actor is destination of the channel?
        if (p->getType() == SDFport::In)
        {
            if (!CH_TOKENS(c->getId(), p->getRate()))
            {
                return false;
            }    
        }
    }

    return true;
}

/**
 * startActorFiring ()
 * Start the actor firing. Remove tokens from all input channels and add the
 * actor firing to the list of active actor firings and advance sequence
 * position.
 */
void SDFstateSpaceListScheduler::TransitionSystem::startActorFiring(
        TimedSDFactor *a)
{
    SDFtime execTime, completionTime, timeTileStartOfSlice, waitingTime;
    int remainingExecTime, nrOfFullRotationsInNonReservedPart;
    
    // Consume tokens from inputs and space for output tokens
    for (SDFportsIter iter = a->portsBegin(); iter != a->portsEnd(); iter++)
    {
        SDFport *p = *iter;
        SDFchannel *c = p->getChannel();
        
        // Actor is destination of the channel?
        if (p->getType() == SDFport::In)
        {
            CONSUME(c->getId(), p->getRate());
        }
    }

    // Execution time of the actor
    execTime = a->getExecutionTime();
    
    // Compute time needed to complete actor firing
    if (archPlatform.actorBinding[a->getId()] == ACTOR_NOT_BOUND)
    {
        completionTime = execTime;
    }
    else
    {
        uint p = archPlatform.actorBinding[a->getId()];
        
        // Time wheel has not yet reached start of slice?
        if (currentState.tdmaPos[p] 
                < archPlatform.tdmaSize[p] - archPlatform.tdmaSlice[p])
        {
            // Wait till start of slice and next complete the number
            // of required wheel rotations to get the full execution time
            // of the slice
            timeTileStartOfSlice = archPlatform.tdmaSize[p] 
                          - archPlatform.tdmaSlice[p] - currentState.tdmaPos[p];
            nrOfFullRotationsInNonReservedPart = (int) ceil((double) execTime 
                          / archPlatform.tdmaSlice[p]) - 1;
            waitingTime = (archPlatform.tdmaSize[p] - archPlatform.tdmaSlice[p])
                          * nrOfFullRotationsInNonReservedPart;
            completionTime = timeTileStartOfSlice + execTime + waitingTime;
        }
        else
        {
            // Start the execution time immediatly, wait the number of wheel
            // rotations after the current slice has been used to complete the
            // firing
            remainingExecTime = (int)execTime - (int)archPlatform.tdmaSize[p] 
                                              + (int)currentState.tdmaPos[p];
            
            if (remainingExecTime < 0)
            {
                waitingTime = 0;
            }
            else
            {
                waitingTime = (remainingExecTime / archPlatform.tdmaSlice[p]) 
                    * (archPlatform.tdmaSize[p] - archPlatform.tdmaSlice[p]);
            }
     
            completionTime = execTime + waitingTime;
        }
        
        // Processor is no longer idle
        procIdle[p] = false;
        
        // Remove actor from the list of ready actors
        actorReadyList[p].pop_front();

        // Add actor to schedule of the processor
        Schedule *s = new Schedule;
        s->actor = a;
        s->last = false;
        s->next = NULL;
        s->previous = NULL;

        // Add actor firing to exisiting schedule
        if (archPlatform.schedule[p] == NULL)
        {
            s->id = 0;
            archPlatform.schedule[p] = s;
        }
        else
        {
            s->id = SCHEDULE_POS(p)->id + 1;
            s->previous = SCHEDULE_POS(p);
            SCHEDULE_POS(p)->next = s;
        }

        // Store last point of schedule in the state
        SCHEDULE_POS(p) = s;
    }

    // Add actor firing to the list of active firings of this actor
    currentState.actClk[a->getId()].push_back(completionTime);
}

/**
 * actorReadyToEnd ()
 * The function returns true when the actor is ready to end its firing. Else
 * the function returns false.
 */
bool SDFstateSpaceListScheduler::TransitionSystem::actorReadyToEnd(SDFactor *a)
{
    if (currentState.actClk[a->getId()].empty())
        return false;
    
    // First actor firing in sorted list has execution time left?
    if (currentState.actClk[a->getId()].front() != 0)
        return false;

    return true;
}

/**
 * endActorFiring ()
 * Produce tokens on all output channels and remove the actor firing from the
 * list of active firings.
 */
void SDFstateSpaceListScheduler::TransitionSystem::endActorFiring(SDFactor *a)
{
    for (SDFportsIter iter = a->portsBegin(); iter != a->portsEnd(); iter++)
    {
        SDFport *p = *iter;
        SDFchannel *c = p->getChannel();
        
        // Actor is source of the channel?
        if (p->getType() == SDFport::Out)
        {
            PRODUCE(c->getId(), p->getRate());
        }
    }

    // Remove the firing from the list of active actor firings
    currentState.actClk[a->getId()].pop_front();

    // Actor bound to processor?
    if (archPlatform.actorBinding[a->getId()] != ACTOR_NOT_BOUND )
    {
        uint p = archPlatform.actorBinding[a->getId()];

        // Processor becomes idle
        procIdle[p] = true;
    }    
}

/**
 * clockStep ()
 * The function progresses time till the first end of firing transition
 * becomes enabled. The time step is returned. In case of deadlock, the
 * time step is equal to UINT_MAX.
 */
SDFtime SDFstateSpaceListScheduler::TransitionSystem::clockStep()
{
    SDFtime step = UINT_MAX;
    
    // Find maximal time progress
    for (uint a = 0; a < appGraph->nrActors(); a++)
    {
        if (!currentState.actClk[a].empty())
        {
            SDFtime actClk = currentState.actClk[a].front();

            if (step > actClk)
                step = actClk;
        }
    }

    // Still actors ready to end their firing?
    if (step == 0)
        return 0;

	// Check for progress (i.e. no deadlock) 
	if (step == UINT_MAX)
        return UINT_MAX;

    // Lower remaining execution time actors
    for (uint a = 0; a < appGraph->nrActors(); a++)
    {
        for (list<SDFtime>::iterator iter = currentState.actClk[a].begin();
                    iter != currentState.actClk[a].end(); iter++)
        {
            SDFtime &actFiringTime = *iter;
            
            // Lower remaining execution time of the actor firing
            actFiringTime -= step;
        }
    }

    // Advance the time wheels
    for (uint t = 0; t < archPlatform.nrTiles; t++)
    {
        currentState.tdmaPos[t] = (currentState.tdmaPos[t] + step) 
                                                    % archPlatform.tdmaSize[t];
    }

    // Advance the global clock  
    currentState.glbClk += step;

    return step;
}

/**  
 * execSDFgraph()  
 * Execute the SDF graph till a deadlock is found or a recurrent state.  
 * The throughput is returned.  
 */  
TDtime SDFstateSpaceListScheduler::TransitionSystem::execSDFgraph()  
{
    StatesIter recurrentState;
    SDFtime clkStep;
    int repCnt = 0;  

    // Clear the list of stored states
    clearStoredStates();

    // Initialize processor states
    actorReadyList.resize(archPlatform.nrTiles);
    procIdle.resize(archPlatform.nrTiles);
    for (uint p = 0; p < archPlatform.nrTiles; p++)
    {
        actorReadyList[p].clear();
        procIdle[p] = true;
    }
    
    // Create initial state
    currentState.init(appGraph->nrActors(), appGraph->nrChannels(),
                        archPlatform.nrTiles);
    currentState.clear();  
    previousState.init(appGraph->nrActors(), appGraph->nrChannels(),
                        archPlatform.nrTiles);
    previousState.clear();

    // Initial tokens and schedules
    for (SDFchannelsIter iter = appGraph->channelsBegin();
            iter != appGraph->channelsEnd(); iter++)
    {
        SDFchannel *c = *iter;
        
        CH(c->getId()) = c->getInitialTokens();
    }
    for (uint p = 0; p < archPlatform.nrTiles; p++)
    {
        SCHEDULE_POS(p) = NULL;
        TDMA_POS(p) = 0;
    }
    
    // Fire the actors  
    while (true)  
    {
        // Finish actor firings  
        for (SDFactorsIter iter = appGraph->actorsBegin();
                iter != appGraph->actorsEnd(); iter++)
        {
            SDFactor *a = *iter;
            
            while (actorReadyToEnd(a))  
            {
                if (outputActor->getId() == a->getId())
                {
                    repCnt++;
                    if (repCnt == outputActorRepCnt)  
                    { 
                        // Add state to hash of visited states  
                        if (!storeState(currentState, recurrentState))
                        {
                            // Complete cycle in each processor schedule
                            for (uint p = 0; p < archPlatform.nrTiles; p++)
                            {
                                if (archPlatform.schedule[p] != NULL)
                                {
                                    // Last schedule should be removed and
                                    // on the one before the last, the 
                                    // 'next' must be set to 'SCHEDULE(p)' in
                                    // recurrent state (so not to SCHEDULE(p)
                                    // itself!) and 'last' must be set to
                                    // 'true'.
                                    if (SCHEDULE_POS(p)->previous != NULL)
                                    {
                                        Schedule *s = SCHEDULE_POS(p)->previous;
                                        s->last = true;
                                        s->next =recurrentState->schedulePos[p];
                                    
                                        // Cleanup recurrent schedule part
                                        delete SCHEDULE_POS(p);
                                    }
                                    else
                                    {
                                        // Single state schedule
                                        Schedule *s = SCHEDULE_POS(p);
                                        s->last = true;
                                        s->next = s;
                                    }
                                }
                            }
                            
                            // Done
                            return computeThroughput(recurrentState);
                        }
	                    currentState.glbClk = 0;
                        repCnt = 0;
                    }  
                }

                // End the actor firing
                endActorFiring(a);
            }  
        }

        // Update the actorReadyList with all firings that became enabled
        // after last clock step
        for (SDFactorsIter iter = appGraph->actorsBegin();
                iter != appGraph->actorsEnd(); iter++)
        {
            TimedSDFactor *a = (TimedSDFactor*)(*iter);

            if (archPlatform.actorBinding[a->getId()] != ACTOR_NOT_BOUND)
            {
                uint t = archPlatform.actorBinding[a->getId()];
                uint nrFiringsCurrent = INT_MAX;
                uint nrFiringsPrevious = INT_MAX;
                uint nrFiringsEnabled;
                
                // Compute number of enabled firings in current 
                // and previous state
                for (SDFportsIter iterP = a->portsBegin();
                        iterP != a->portsEnd(); iterP++)
                {
                    SDFport *p = *iterP;
                    SDFchannel *c = p->getChannel();
                    uint nrCurrent, nrPrevious;
                    
                    if (p->getType() == SDFport::In)
                    {
                        nrCurrent = currentState.ch[c->getId()] / p->getRate();
                        nrPrevious = previousState.ch[c->getId()] /p->getRate();
                        
                        if (nrCurrent < nrFiringsCurrent)
                            nrFiringsCurrent = nrCurrent;
                        if (nrPrevious < nrFiringsPrevious)
                            nrFiringsPrevious = nrPrevious;
                    }
                }
                
                if (nrFiringsCurrent <= nrFiringsPrevious)
                    nrFiringsEnabled = 0;
                else
                    nrFiringsEnabled = nrFiringsCurrent - nrFiringsPrevious;
                
                // Add as many firings to the actorReadyList as became enabled
                for (uint i = 0; i < nrFiringsEnabled; i++)
                    actorReadyList[t].push_back(a);
            }
        }        

        // Start actor firings  
        for (SDFactorsIter iter = appGraph->actorsBegin();
                iter != appGraph->actorsEnd(); iter++)
        {
            TimedSDFactor *a = (TimedSDFactor*)(*iter);
        
            // Ready to fire actor a?
            while (actorReadyToFire(a))
            {
                // Fire actor a
                startActorFiring(a);
            }
        }

        // Clock step
        clkStep = clockStep();

        // Deadlocked?
        if (clkStep == UINT_MAX)
        {
            return 0;
        }

        // Store partial state to check for progress
        for (uint i = 0; i < appGraph->nrChannels(); i++)
        {
            previousState.ch[i] = currentState.ch[i];
        }
    }
    
    return 0;
}  

/**
 * List scheduler
 * Constructs a set of static-order schedules for the processors in the
 * architecture platform. Each schedule orders the firings of the actors
 * running on the processor. A list scheduler is used to construct the
 * static-order schedules.
 */
void SDFstateSpaceListScheduler::schedule(TimedSDFgraph *gr, PlatformGraph *ap)
{
    // Check that the application graph is a strongly connected graph
    if (!isStronglyConnectedGraph(gr))
        throw CException("Graph is not strongly connected.");

    // Create a transition system
    TransitionSystem transitionSystem(gr, ap);

    // Execute the graph to construct a schedule
    transitionSystem.execSDFgraph();

    // Assign schedules to processors in the tiles
    for (uint i = 0; i < transitionSystem.archPlatform.nrTiles; i++)
    {
        // Schedule constructed for processor i?
        if (transitionSystem.archPlatform.schedule[i] != NULL)
        {
            Tile *t = ap->getTile(i);
            Processor *p = t->getProcessor();
            
            p->setSchedule(transitionSystem.archPlatform.schedule[i]);
        }
    }
}
