/*
 *  TU Eindhoven
 *  Eindhoven, The Netherlands
 *
 *  Name            :   schedule.cc
 *
 *  Author          :   Sander Stuijk (sander@ics.ele.tue.nl)
 *
 *  Date            :   July 27, 2005
 *
 *  Function        :   Schedule for SDF graphs
 *
 *  History         :
 *      27-07-05    :   Initial version.
 *
 * $Id: schedule.cc,v 1.1.1.1 2007/10/02 10:59:47 sander Exp $
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * 
 * In other words, you are welcome to use, share and improve this program.
 * You are forbidden to forbid anyone else to use, share and improve
 * what you give them.   Happy coding!
 */

#include "schedule.h"

/**
 * ~SPschedule ()
 * Destructor.
 */
SPschedule::~SPschedule()
{
    for (SPschedulesIter iter = subScheduleBegin();
            iter != subScheduleEnd(); iter++)
    {
        SPschedule *s = *iter;

        delete s;
    }
}

/**
 * addActor ()
 * Add a sub-schedule to this schedule containing one actor
 */
void SPschedule::addActor(const CId actorId, const uint repCnt)
{
    SPschedule *s = new SPschedule(sdfGraph);
    s->repCnt = repCnt;
    s->actorId = actorId;
    
    subSchedule.push_back(s);
}

/**
 * ~MPschedule ()
 * Destructor.
 */
MPschedule::~MPschedule()
{
    for (SPschedulesIter iter = schedulesBegin();
            iter != schedulesEnd(); iter++)
    {
        SPschedule *s = *iter;

        delete s;
    }
}
/**
 * getSPschedule ()
 * Reference to the single-processor schedule with the given id.
 */
SPschedule *MPschedule::getSPschedule(const CId id)
{
    for (SPschedulesIter iter = spSchedules.begin();
            iter != spSchedules.end(); iter++)
    {
        SPschedule *s = *iter;
        
        if (s->getId() == id)
            return s;
    }
    
    throw CException("SPschedule does not exist in MPschedule.");
}

/**
 * addSPschedule ()
 * Create a new single processor schedule and make it equavalent
 * to the supplied schedule s.
 */
void MPschedule::addSPschedule(SPschedule *s)
{
    // Add schedule to list of schedules
    spSchedules.push_back(s);
}

/**
 * newSPschedule ()
 * Create a new single-processor schedule and return a reference to it.
 */
SPschedule *MPschedule::newSPschedule(TimedSDFgraph *g, const CId id)
{
    SPschedule *s = new SPschedule(g, id);
    spSchedules.push_back(s);
    
    return spSchedules.back();
}

/**
 * removeSPschedule ()
 * Remove the single-processor schedule with the given id from the set
 * of schedules.
 */
void MPschedule::removeSPschedule(const CId id)
{
    for (SPschedulesIter iter = spSchedules.begin();
            iter != spSchedules.end(); iter++)
    {
        SPschedule *s = *iter;
        
        if (s->getId() == id)
        {
            delete *iter;
            spSchedules.erase(iter);
            return;
        }
    }
    
    throw CException("SPschedule does not exist in MPschedule.");    
}

/**
 * print ()
 * Print single-processor schedule of graph g to output stream.
 */
void SPschedule::print(ostream &out)
{
    if (repCnt != 1)
        out << repCnt;

    if (isSubSchedule())
    {
        out << "(";
        for (SPschedulesIter iter = subScheduleBegin();
                iter != subScheduleEnd(); iter++)
        {
            SPschedule *s = *iter;
            s->print(out);    
        }
        out << ")";
    }
    else
    {
        out << sdfGraph->getActor(actorId)->getName() << ";";
    }    
}

/**
 * print ()
 * Print multi-processor schedule of graph g to output stream.
 */
void MPschedule::print(ostream &out)
{
    for (SPschedulesIter iter = schedulesBegin();
            iter != schedulesEnd(); iter++)
    {
        SPschedule *s = *iter;
        
        out << "Schedule(" << s->getId() << "):" << endl;
        s->print(out);
        out << endl;
    }
}
